(in-package :cl-uxn)

(defvar *breakpoints* ()
  "List of addresses which the cl-uxn:run function should invoke the signal uxn-breakpoint.")
(defvar *watchpoints* ())
(defvar *device-input* nil
  "Callback function with two arguments: The port written to and the value.")
(defvar *device-output* nil
  "Callback function with one argument: The port read from. It must return an integer.")

(define-condition uxn-breakpoint (uxn-condition)
  ())
(define-condition uxn-watchpoint (uxn-condition)
  ((address :initarg :address :reader address)
   (old-value :initarg :old-value :reader old-value)
   (new-value :initarg :new-value :reader new-value)))

(defstruct uxn
  (memory (make-array #x10000
                      :element-type 'uint8
                      :initial-element 0)
   :type (simple-array uint8 (#x10000))
   :read-only t)
  (work-stack (make-uxn-stack)
   :type uxn-stack
   :read-only t)
  (return-stack (make-uxn-stack)
   :type uxn-stack
   :read-only t)
  (program-counter 0
   :type uint16))

(defstruct uxn-stack
  (buffer (make-array #x100
                      :element-type 'uint8
                      :initial-element 0)
   :type (simple-array uint8 (#x100))
   :read-only t)
  (index #xff
   :type uint8))

(defun mode-short-p (opcode) (logbitp 5 opcode))
(defun mode-return-p (opcode) (logbitp 6 opcode))
(defun mode-keep-p (opcode) (logbitp 7 opcode))
(defun stack-pop (stack wide)
  (let ((result (aref (uxn-stack-buffer stack) (uxn-stack-index stack))))
    (setf-8 (uxn-stack-index stack) (1- (uxn-stack-index stack)))
    (when wide
      (setf (ldb (byte 8 8) result)
            (aref (uxn-stack-buffer stack) (uxn-stack-index stack)))
      (setf-8 (uxn-stack-index stack)
              (1- (uxn-stack-index stack))))
    result))
(defun stack-push (stack value wide)
  (let ((result (mod value #x10000)))
    (setf-8 (uxn-stack-index stack)
            (1+ (uxn-stack-index stack)))
    (when wide
      (setf (aref (uxn-stack-buffer stack) (uxn-stack-index stack))
            (ldb (byte 8 8) result))
      (setf-8 (uxn-stack-index stack)
              (1+ (uxn-stack-index stack))))
    (setf (aref (uxn-stack-buffer stack) (uxn-stack-index stack))
          (ldb (byte 8 0) result))))
(defun memory-read (memory address chop wide)
  (let ((result (aref memory address)))
    (when wide
      (setf result (ash result 8))
      (setf (ldb (byte 8 0) result) (aref memory (mod (1+ address)
                                                                   chop))))
    result))
(defun memory-write (memory address value chop wide)
  (let ((result (mod value #x10000)))
    (cond (wide (setf-8 (aref memory address)
                        (ldb (byte 8 8) result))
                (setf-8 (aref memory (mod (1+ address)
                                                       chop))
                        (ldb (byte 8 0) result)))
          (t (setf-8 (aref memory address)
                     (ldb (byte 8 0) result))))))

(defmacro with-uxn ((state) &body body)
  `(with-accessors ((memory uxn-memory)
                    (work-stack uxn-work-stack)
                    (return-stack uxn-return-stack)
                    (program-counter uxn-program-counter))
       ,state
     ,@body))

(defparameter *opcode-dispatch-table*
  (make-array #x100
              :element-type 'function
              :initial-element (lambda (instance opcode)
                                 (declare (ignore instance opcode))
                                 (error "Unknown opcode"))))

;;; General uxn functions

(defun make-uxn-with-image (initial-memory)
  "Creates an instance with a starting memory array."
  (assert (<= (length initial-memory) #xff00))
  (let ((instance (make-uxn)))
    (replace (uxn-memory instance) initial-memory :start1 #x100)
    instance))

(defun run (instance &optional vector)
  "Run the routine at address vector in the specified instance."
  (with-uxn (instance)
    (when vector (setf program-counter vector))
    (do ((returned nil)) (returned)
      (when (member program-counter *breakpoints*)
        (signal 'uxn-breakpoint instance))
      (setf returned (run-step instance)))))

(defun run-step (instance)
  "Execute the instruction at the instance's program counter."
  (with-uxn (instance)
    (let ((opcode (aref memory program-counter)))
      (setf-16 program-counter (1+ program-counter))
      (funcall (aref *opcode-dispatch-table* opcode) instance opcode)
      (zerop opcode))))
